#!/usr/bin/env python
from __future__ import absolute_import, unicode_literals
if __name__ == '__main__':
    import os
    import sys
    sys.path.insert(0, os.path.dirname(os.path.dirname(
        os.path.dirname(os.path.abspath(__file__))
    )))

import unittest
from autumn import validators
from autumn import utils
from autumn.connections import get_db
from autumn.models import Model, QS, qn
from autumn.tests.models import Book, Author

class TestUtils(unittest.TestCase):
    
    maxDiff = None

    def test_resolve(self):
        from autumn.connections import DummyCtx
        self.assertTrue(utils.resolve('autumn.connections.DummyCtx') is DummyCtx)


class TestModels(unittest.TestCase):

    maxDiff = None

    def testmodel(self):
        # Create tables

        create_sql = {
            'postgresql': """
                DROP TABLE IF EXISTS autumn_tests_models_author;
                CREATE TABLE autumn_tests_models_author (
                    id serial NOT NULL PRIMARY KEY,
                    first_name VARCHAR(40) NOT NULL,
                    last_name VARCHAR(40) NOT NULL,
                    bio TEXT
                );
                DROP TABLE IF EXISTS books;
                CREATE TABLE books (
                    id serial NOT NULL PRIMARY KEY,
                    title VARCHAR(255),
                    author_id integer REFERENCES autumn_tests_models_author(id) ON DELETE CASCADE
                );
             """,
            'mysql': """
                DROP TABLE IF EXISTS autumn_tests_models_author;
                CREATE TABLE autumn_tests_models_author (
                    id INT(11) NOT NULL auto_increment,
                    first_name VARCHAR(40) NOT NULL,
                    last_name VARCHAR(40) NOT NULL,
                    bio TEXT,
                    PRIMARY KEY (id)
                );
                DROP TABLE IF EXISTS books;
                CREATE TABLE books (
                    id INT(11) NOT NULL auto_increment,
                    title VARCHAR(255),
                    author_id INT(11),
                    FOREIGN KEY (author_id) REFERENCES autumn_tests_models_author(id),
                    PRIMARY KEY (id)
                );
             """,
            'sqlite3': """
                DROP TABLE IF EXISTS autumn_tests_models_author;
                DROP TABLE IF EXISTS books;
                CREATE TABLE autumn_tests_models_author (
                  id INTEGER PRIMARY KEY AUTOINCREMENT,
                  first_name VARCHAR(40) NOT NULL,
                  last_name VARCHAR(40) NOT NULL,
                  bio TEXT
                );
                CREATE TABLE books (
                  id INTEGER PRIMARY KEY AUTOINCREMENT,
                  title VARCHAR(255),
                  author_id INT(11),
                  FOREIGN KEY (author_id) REFERENCES autumn_tests_models_author(id)
                );
            """
        }
        for table in ('autumn_tests_models_author', 'books'):
            get_db().execute('DELETE FROM {0}'.format(qn(table)))

        # Test Creation
        james = Author(first_name='James', last_name='Joyce')
        james.save()

        kurt = Author(first_name='Kurt', last_name='Vonnegut')
        kurt.save()

        tom = Author(first_name='Tom', last_name='Robbins')
        tom.save()

        Book(title='Ulysses', author_id=james.id).save()
        Book(title='Slaughter-House Five', author_id=kurt.id).save()
        Book(title='Jitterbug Perfume', author_id=tom.id).save()
        slww = Book(title='Still Life with Woodpecker', author_id=tom.id)
        slww.save()

        # Test Model.pk getter and setter
        pk = slww.pk
        self.assertEqual(slww.pk, slww.id)
        slww.pk = tom.id
        self.assertEqual(slww.pk, tom.id)
        slww.pk = pk
        self.assertEqual(slww.pk, pk)

        self.assertTrue(kurt == Author.get(kurt.id))
        self.assertTrue(kurt != tom)
        self.assertEqual(hash(kurt), kurt.pk)

        # Test ForeignKey
        self.assertEqual(slww.author.first_name, 'Tom')
        slww.author = kurt
        self.assertEqual(slww.author.first_name, 'Kurt')
        del slww.author
        self.assertEqual(slww.author, None)
        slww.author = None
        self.assertEqual(slww.author, None)
        slww.author = tom.pk
        self.assertEqual(slww.author.first_name, 'Tom')

        # Test OneToMany
        self.assertEqual(len(list(tom.books)), 2)

        kid = kurt.id
        del(james, kurt, tom, slww)

        # Test retrieval
        b = Book.get(title='Ulysses')[0]

        a = Author.get(id=b.author_id)[0]
        self.assertEqual(a.id, b.author_id)

        a = Author.get(id=b.id)[:]
        #self.assert_(isinstance(a, list))
        self.assert_(isinstance(a, QS))

        # Test update
        new_last_name = 'Vonnegut, Jr.'
        a = Author.get(id=kid)[0]
        a.last_name = new_last_name
        a.save()

        a = Author.get(kid)
        self.assertEqual(a.last_name, new_last_name)

        # Test count
        self.assertEqual(Author.get().count(), 3)

        self.assertEqual(len(Book.get()[1:4]), 3)

        # Test delete
        a.delete()
        self.assertEqual(Author.get().count(), 2)

        # Test validation
        a = Author(first_name='', last_name='Ted')
        try:
            a.save()
            raise Exception('Validation not caught')
        except Model.ValidationError:
            pass

        # Test defaults
        a.first_name = 'Bill and'
        a.save()
        self.assertEqual(a.bio, 'No bio available')

        try:
            Author(first_name='I am a', last_name='BadGuy!').save()
            raise Exception('Validation not caught')
        except Model.ValidationError:
            pass

        print '### Testing for smartsql integration'
        t = Author.s
        fields = [t.qs.sqlrepr(i) for i in t.get_fields()]
        if get_db().engine == 'postgresql':
            self.assertListEqual(
                fields,
                ['"autumn_tests_models_author"."id"',
                 '"autumn_tests_models_author"."first_name"',
                 '"autumn_tests_models_author"."last_name"',
                 '"autumn_tests_models_author"."bio"' ]
            )
        else:
            self.assertListEqual(
                fields,
                ['`autumn_tests_models_author`.`id`',
                 '`autumn_tests_models_author`.`first_name`',
                 '`autumn_tests_models_author`.`last_name`',
                 '`autumn_tests_models_author`.`bio`', ]
            )

        if get_db().engine == 'postgresql':
            self.assertEqual(Book.s.author, '"book"."author_id"')
        else:
            self.assertEqual(Book.s.author, '`book`.`author_id`')

        qs = t.qs
        if get_db().engine == 'postgresql':
            self.assertEqual(qs.sqlrepr(), '''SELECT "autumn_tests_models_author"."id", "autumn_tests_models_author"."first_name", "autumn_tests_models_author"."last_name", "autumn_tests_models_author"."bio" FROM "autumn_tests_models_author"''')
        else:
            self.assertEqual(qs.sqlrepr(), """SELECT `autumn_tests_models_author`.`id`, `autumn_tests_models_author`.`first_name`, `autumn_tests_models_author`.`last_name`, `autumn_tests_models_author`.`bio` FROM `autumn_tests_models_author`""")
        self.assertEqual(len(qs), 3)
        for obj in qs:
            self.assertTrue(isinstance(obj, Author))

        qs = qs.where(t.id == b.author_id)
        if get_db().engine == 'postgresql':
            self.assertEqual(qs.sqlrepr(), """SELECT "autumn_tests_models_author"."id", "autumn_tests_models_author"."first_name", "autumn_tests_models_author"."last_name", "autumn_tests_models_author"."bio" FROM "autumn_tests_models_author" WHERE ("autumn_tests_models_author"."id" = %s)""")
        else:
            self.assertEqual(qs.sqlrepr(), """SELECT `autumn_tests_models_author`.`id`, `autumn_tests_models_author`.`first_name`, `autumn_tests_models_author`.`last_name`, `autumn_tests_models_author`.`bio` FROM `autumn_tests_models_author` WHERE (`autumn_tests_models_author`.`id` = %s)""")
        self.assertEqual(len(qs), 1)
        self.assertTrue(isinstance(qs[0], Author))

    def testvalidators(self):
        ev = validators.Email()
        assert ev('test@example.com')
        assert not ev('adsf@.asdf.asdf')
        assert validators.Length()('a')
        assert not validators.Length(2)('a')
        assert validators.Length(max_length=10)('abcdegf')
        assert not validators.Length(max_length=3)('abcdegf')

        n = validators.Number(1, 5)
        assert n(2)
        assert not n(6)
        assert validators.Number(1)(100.0)
        assert not validators.Number()('rawr!')

        vc = validators.ValidatorChain(validators.Length(8), validators.Email())
        assert vc('test@example.com')
        assert not vc('a@a.com')
        assert not vc('asdfasdfasdfasdfasdf')

if __name__ == '__main__':
    unittest.main()
